#include "valout.h"

#ifdef __AVR__
#include <avr/pgmspace.h>

#ifdef VALOUT_32
#define TABLE_READ(ADDR) pgm_read_dword (ADDR)
#else
#define TABLE_READ(ADDR) pgm_read_word (ADDR)
#endif
#else
#define PROGMEM
#define TABLE_READ (*(p))
#endif // avr-gcc

#ifndef VALOUT_DECIMAL_POINT
#define VALOUT_DECIMAL_POINT '.'
#endif // VALOUT_DECIMAL_POINT

// maximum number length
// 2.40823 = log 256 / log 10
static char buff[(int) 2.40823*sizeof(valout_data_t)+1 + sizeof ("-.")];

static const valout_data_t TEST[] PROGMEM =
{
		 10L
		,100L
		,1000L
		,10000L
#ifdef VALOUT_32
		,100000L
		,1000000L
		,10000000L
		,100000000L
		,1000000000L
#endif // VALOUT_32
};

#define TABLE_SIZE (sizeof(TEST) / sizeof (TEST[0]))

void valout (valout_data_t val, int8_t digits, int8_t point, char * outp)
{
	uint8_t notzero = 0;
	char * buffp = buff;
	
	// handle sign
	
	if (val < 0)
		{
			*outp++ = '-';
			val = -val;
		}
	
	point -= TABLE_SIZE + 1;
	
	for (uint8_t i = TABLE_SIZE; i; i--)
		{
			char d = '0'-1;
			
			// power_10_i = 10 ** (i+1)
			valout_data_t power_10_i = (valout_data_t) TABLE_READ (& TEST[i-1]);
			
			// get i-th digit
			do
				d++, val -= power_10_i;
			while (val >= 0);
			val += power_10_i;
			
			if (val >= 0)
			notzero++;
			
			if (notzero)
			*buffp++ = d;
			
			// insert decimal point
			if (++point == 0)
			{
				// zero prior decimal point
				if (notzero == 0)
					*buffp++ = d;
		
					notzero++;
					
					*buffp++ = VALOUT_DECIMAL_POINT;
			}
		}

	// remainder = ones
	*buffp++ = val + '0';
	*buffp = '\0';
	digits -= ((int) buffp) - ((int) buff); // length of string

	// insert spaces until min length
	while( --digits >= 0 )
		*outp++ = ' ';
		
	buffp = buff;

	// always append zero
	do
		*outp++ = *buffp;
		while (*buffp++);
}
